/**********************************************
* wmmsg.c
*
* Maintains an X window that displays incoming
* message events. Listens for new events by
* listening on a UNIX socket.
**********************************************/
#include "wmmsg.h"
#include "wmconfig.h"
#include "wmmsg.xpm"

/*************
* fun headers
*************/
static int process_msg (int);
static int getServerSocket (char*);
static int addMessage (struct message*);
static int removeMessage (int);
static void removeProtocol(int);
static void removeAllMessages ();
static void printAllMessages ();
static void BlitString (char *, int, int);
static int get_desktop_num (void);
static char * getSocketName ();
static void clean_up (int);
static void displayUsage (void);
static void outOfMemory(void);

/*********************
* classwide variables
*********************/
//global configuration structure
static struct config_data *global_config;
static char global_wmmsg_mask_bits[64 * 64];
//name of socket we're listening on
static char *global_strSocketName;
//all the messages we're listening to, implemented as a stack
static struct message *global_messages;
//slot the last message is located
static int global_intLastMessage = -1;
//does the screen need to be updated?
static short global_screen_needs_updating = 0;
//number of blinks left per protocol (if grouped
//icons) or display slot (if not grouped)
static int *global_blinks_left;
//X shit for getting desktop
static Atom global_win_workspace;

/***************************
* long command line options
***************************/
#ifdef HAVE_GETOPT_LONG
static struct option long_options[] =
{
    {   "help",       no_argument, NULL, 'h'},
    {"version",       no_argument, NULL, 'V'},
    {"verbose",       no_argument, NULL, 'v'},
    { "config", required_argument, NULL, 'c'},
    {0, 0, 0, 0}
};
#else
void *long_options = NULL;
#endif

/******
* main
******/
int main (int argc, char *argv[])
{
    //file descriptors for sockets
    int fdSocket, fdClient;
    //for socket listening
    fd_set socks;
    struct timeval listen_timeout;
    int socketHasMessage;
    //X event handling stuff
    XEvent Event;
    //location of icons on screen
    int intDest_y;
    //boolean list of which protocols are printed
    short *protocols_displayed = 0;
    //the protocols that each slot displays in the gui (-1 if blank)
    int protocol_at_pos[MESSAGES_PER_SCREEN];
    //number of icons currently being displayed
    int intIconsDisplayed;
    //for keeping track of desktop
    int intDesktop = -1, intPrevDesktop;
    //temps
    int intCurrOpt, intMsgNo, intBlinkIndex=0, clear_blink;
    gboolean gtk_initialized;
    struct message *msgCurr;
    int i, j;
    char *home_dir=0, *path_to_rc=0;

    //initialize gtk
    gtk_initialized = gtk_init_check (&argc, &argv);

    /********************
    * parse command line
    ********************/
	while ((intCurrOpt = wmmsg_getopt(argc, argv, "hVvc:",
        (void *)long_options, NULL)) != -1)
    {
        switch(intCurrOpt)
        {
            //-v: verbose mode
            case 'v':
                verbose = 1;
                break;
            //-V: version
            case 'V':
                printf ("Version: %s\n", VERSION);
                _exit (0);
            //-c: config file
            case 'c':
                if(verbose) fprintf(stderr, "using config file \"%s\"\n", optarg);
                path_to_rc = malloc(sizeof(char) * strlen(optarg)+1);
                if(!path_to_rc) outOfMemory();
                strncpy(path_to_rc, optarg, strlen(optarg)+1);
                break;
            //-h: help
            case 'h':
                displayUsage();
                _exit (0);
            //other options: error
            case ':':
                fprintf(stderr, "Option -c requires a file name\n");
            case '?':
                //getopt already printed an error message
                fprintf(stderr, "Try `wmmsg -h' for more information.\n");
                break;
        }
    }

    //if not able to open a display, then err
    if(gtk_initialized == 0)
    {
        fprintf(stderr, "Sorry, there was a GTK error initializing your display. Exiting.\n");
        _exit(1);
    }

    //draw initial picture
    createXBMfromXPM (global_wmmsg_mask_bits, wmmsg_xpm,
        WMMSG_MASK_WIDTH, WMMSG_MASK_HEIGHT);
    openXwindow (argc, argv, wmmsg_xpm, global_wmmsg_mask_bits,
        WMMSG_MASK_WIDTH, WMMSG_MASK_HEIGHT);
    global_win_workspace = XInternAtom(display, "_NET_CURRENT_DESKTOP", 0);

    //add clickable regions
    for(i = 0; i < MESSAGES_PER_SCREEN; i++)
        AddMouseRegion (i,  6, 6+i*13, 51, (6+i*13) + 13);

    /******************************
    * import data from config file
    ******************************/
    if(verbose && (display == NULL)) fprintf(stderr, "display is NULL!!!\n");
    if(path_to_rc == NULL)
    {
        home_dir = getenv("HOME");
        if(home_dir)
        {
            path_to_rc = malloc(sizeof(char)*strlen(home_dir)+12);
            if(!path_to_rc) outOfMemory();
            strncpy(path_to_rc, home_dir, strlen(home_dir)+1);
            strncat(path_to_rc, "/.wmmsgrc", 10);
        }
        else
        {
            path_to_rc = malloc(sizeof(char)*14);
            if(!path_to_rc) outOfMemory();
            strncpy(path_to_rc, "/etc/wmmsgrc", 13);
        }
    }
    global_config = get_new_config_data(path_to_rc);
    free(path_to_rc);

    global_messages = malloc(sizeof(struct message)*MAX_MESSAGES);
    if(!global_messages) outOfMemory();
    protocols_displayed = calloc(PROTOCOL_COUNT, sizeof(short));
    if(!protocols_displayed) outOfMemory();

    if(BLINK_ENABLED)
    {
        if(GROUPED_ICONS)
            global_blinks_left = calloc(PROTOCOL_COUNT, sizeof(int));
        else
            global_blinks_left = calloc(MESSAGES_PER_SCREEN, sizeof(int));

        if(!global_blinks_left) outOfMemory();
    }

    //initialize some data
    for(i=0; i<MESSAGES_PER_SCREEN; i++)
        protocol_at_pos[i] = -1;
    if(MONITOR_DESKTOPS)
        intDesktop = get_desktop_num();

    /*******************
    * initialize socket
    *******************/
    //set socket name
    global_strSocketName = getSocketName();

    //handle interrupt events
    signal(SIGINT,  clean_up);
    signal(SIGALRM, clean_up);
    signal(SIGHUP,  clean_up);
    signal(SIGQUIT, clean_up);
    signal(SIGTERM, clean_up);

    //delete socket, in case it exists
    unlink (global_strSocketName);

    //initially open listening socket
    fdSocket = getServerSocket(global_strSocketName);
    if(fdSocket == -1)
    {
        perror ("Error initializing socket");
        _exit(1);
    }

    /***************
    * infinite loop
    ***************/
    while(1)
    {
        //prevents zombie processes from forks
        waitpid(0, NULL, WNOHANG);

        /***********************
        * handle desktop switch
        ***********************/
        if(MONITOR_DESKTOPS)
        {
            intPrevDesktop = intDesktop;
            intDesktop = get_desktop_num();
            if((intDesktop > -1) && (intDesktop != intPrevDesktop))
            {
                //clear all messages
                if(intDesktop == MSG_DESKTOP_NUM)
                {
                    removeAllMessages();
                    global_screen_needs_updating = 1;
                }
            }
        }

        /*************************************************
        * update window before redrawing it, if necessary
        *************************************************/
        if(global_screen_needs_updating)
        {
            //zero out variables
            global_screen_needs_updating = 0;
            intMsgNo = 0;
            intIconsDisplayed = 0;
            for(i = 0; i < PROTOCOL_COUNT; i++)
                protocols_displayed[i] = 0;

            //while we still have room on the screen, cycle
            //through all msgs and display them if necessary
            while((intIconsDisplayed < MESSAGES_PER_SCREEN) && (intMsgNo <= global_intLastMessage))
            {
                //get next message
                msgCurr = &(global_messages[intMsgNo]);

                //determine where to look up blink count
                if(BLINK_ENABLED)
                {
                    if(GROUPED_ICONS)
                        intBlinkIndex = msgCurr->protocol;
                    else
                        intBlinkIndex = intIconsDisplayed;
                }

                //display current message unless we already
                //took care of its protocol group
                if((!GROUPED_ICONS) ||
                    (protocols_displayed[msgCurr->protocol] == 0))
                {
                    intDest_y = 6 + intIconsDisplayed*13;

                    //show the data unless blinking
                    if((!BLINK_ENABLED) ||
                        (global_blinks_left[intBlinkIndex] <= 0) ||
                        ((global_blinks_left[intBlinkIndex] % 2) == 1))
                    {
                        //debug shit
                        if(verbose)
                        {
                            for(i=0; i<intMsgNo; i++)
                                fprintf(stderr, "  ");

                            if(BLINK_ENABLED)
                                fprintf(stderr, "showing [%i, slot %i]  (blinks: %i,"
                                    " blinkIndex: %i)\n", intMsgNo, intIconsDisplayed,
                                    global_blinks_left[intBlinkIndex], intBlinkIndex);
                            else
                                fprintf(stderr, "showing [%i, slot %i]\n",
                                    intMsgNo, intIconsDisplayed);
                        }

                        //display icon from the pixmap
                        XCopyArea (display,
                            global_config->protocol_pixmaps[msgCurr->protocol],
                            wmgen.pixmap, NormalGC, 0, 0, 14, 13, 6, intDest_y);
                        //and the time, too
                        BlitString(msgCurr->time, 21, intDest_y+2);
                    }
                    //hide data if blinking
                    else
                    {
                        if(verbose) for(i=0; i<intMsgNo; i++)
                            fprintf(stderr, "  ");
                        if(verbose) fprintf(stderr,
                            "hiding [%i, slot %i]  (blinks: %i)\n",
                            intMsgNo, intIconsDisplayed,
                            global_blinks_left[intBlinkIndex]);
                        copyXPMArea (BLANK_XPM_X, BLANK_XPM_Y, 46, 13, 6, intDest_y);
                    }

                    //decrement current blinking msg
                    if((BLINK_ENABLED) && (global_blinks_left[intBlinkIndex] > 0))
                    {
                        global_screen_needs_updating = 1;
                        global_blinks_left[intBlinkIndex]--;
                    }

                    if(verbose) fprintf(stderr, "protocol_at_pos[%i] = %i\n",
                        intIconsDisplayed, msgCurr->protocol);
                    protocol_at_pos[intIconsDisplayed] = msgCurr->protocol;

                    protocols_displayed[msgCurr->protocol] = 1;
                    intIconsDisplayed++;
                }
                intMsgNo++;
            }

            //clear rest of space if necessary
            while(intIconsDisplayed < MESSAGES_PER_SCREEN)
            {
                if(protocol_at_pos[intIconsDisplayed] > -1)
                {
                    intDest_y = 6 + intIconsDisplayed*13;
                    copyXPMArea (BLANK_XPM_X, BLANK_XPM_Y, 46, 13, 6, intDest_y);
                    protocol_at_pos[intIconsDisplayed] = -1;
                    if(BLINK_ENABLED && (!GROUPED_ICONS))
                    {
                        if(verbose) fprintf(stderr, "clearing blinks_left[%i]\n", intIconsDisplayed);
                        global_blinks_left[intIconsDisplayed] = 0;
                    }
                }
                intIconsDisplayed++;
            }

            //zero out blink count for protocols that aren't displayed
            //only need to do this if grouped_icons is on, cuz otherwise
            //we're only keeping track of the blinks for each slot
            if(BLINK_ENABLED && GROUPED_ICONS)
            {
                //for each protocol...
                for(i=0; i<PROTOCOL_COUNT; i++)
                {
                    //if protocol has blinks left, clear global_blinks_left[]
                    //unless it's already being displayed on the screen
                    if(global_blinks_left[i] > 0)
                    {
                        clear_blink = 1;
                        for(j=0; j<MESSAGES_PER_SCREEN; j++)
                        {
                            if(protocol_at_pos[j] == i)
                            {
                                clear_blink = 0;
                                break;
                            }
                        }

                        if(clear_blink)
                        {
                            if(verbose) fprintf(stderr, "clearing blinks_left[%i]\n", i);
                            global_blinks_left[i] = 0;
                        }
                    }
                }
            }

        }

        RedrawWindow();

        /*******************************
        * check the socket for messages
        *******************************/
        FD_ZERO(&socks);
        FD_SET(fdSocket, &socks);
        listen_timeout.tv_sec = LISTEN_TIMEOUT_SECS; //sex
        listen_timeout.tv_usec = LISTEN_TIMEOUT_USECS; //microsex
        socketHasMessage = select(fdSocket+1, &socks,
            (fd_set *) 0, (fd_set *) 0, &listen_timeout);

        /*********************************************
        * if there is an incoming message, process it
        *********************************************/
        if(socketHasMessage > 0)
        {
            //accept socket...
            fdClient = accept(fdSocket, 0, 0);
            if(fdClient == -1)
            {
                perror ("Error accepting socket connection");
                _exit(1);
            }
            //process message...
            else if (process_msg(fdClient) == -1)
            {
                perror ("Error processing socket message");
                _exit(1);
            }
            //close socket.
            else if (close(fdClient) == -1)
            {
                perror ("Error closing socket");
                _exit(1);
            }
        }
        else if(socketHasMessage < 0)
        {
            perror ("Error listening to socket connection");
            _exit(1);
        }
        //done with socket stuff

        /*****************
        * handle X events
        *****************/
        while(XPending(display))
        {
            XNextEvent(display, &Event);
            switch (Event.type)
            {
                case Expose:
                    RedrawWindow();
                    break;
                case DestroyNotify:
                    XCloseDisplay(display);
                    _exit (0);
                /***************************
                * user clicked on something
                ***************************/
                case ButtonPress:
                    i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
                    //if clicked on one of the regions
                    if (i >= 0)
                    {
                        //clear all if specified
                        if(CLICK_CLEARS_ALL)
                        {
                            removeAllMessages();
                            //update screen on next refresh
                            global_screen_needs_updating = 1;
                        }
                        //else if there's an icon there, remove it from the list
                        else if(protocol_at_pos[i] > -1)
                        {
                            if(GROUPED_ICONS)
                                removeProtocol(protocol_at_pos[i]);
                            else
                                removeMessage(i);
                            //update screen on next refresh
                            global_screen_needs_updating = 1;
                        }
                    }
                    break;
                case ButtonRelease:
                    RedrawWindow ();
                    break;
            } //end switch Event.type
        } //end X events

    }
    //(not reached)

} //end main

/*************************
* gives command line help
*************************/
void displayUsage(void)
{
#ifdef HAVE_GETOPT_LONG
    fprintf (stderr, "Usage: wmmsg [options]\n\n");
    fprintf (stderr, "Options:\n");
    fprintf (stderr, "    -h, --help          display this text and exit\n"); 
    fprintf (stderr, "    -v, --verbose       verbose output\n");
    fprintf (stderr, "    -V, --version       print version number and exit\n");
    fprintf (stderr, "    -c, --config        use a specified config file\n");
    fprintf (stderr, "\n");
#else
    fprintf (stderr, "Usage: wmmsg [options]\n\n");
    fprintf (stderr, "Options:\n");
    fprintf (stderr, "    -h,       display this text and exit\n"); 
    fprintf (stderr, "    -v,       verbose output\n");
    fprintf (stderr, "    -V,       print version number and exit\n");
    fprintf (stderr, "    -c,       use a specified config file\n");
    fprintf (stderr, "\n");
#endif
    return;
}

/***********************************************
* opens and starts listening on a UNIX socket
*
* in: name of socket
* out: file descriptor of the socket. -1 if err
***********************************************/
int getServerSocket(char p_strSocketName[])
{
    //gotta use this
    struct sockaddr_un name;
    //file descriptor we're using
    int fdCurrSock;

    name.sun_family = AF_UNIX;
    strncpy (name.sun_path, p_strSocketName, sizeof(name.sun_path) - 1);
    name.sun_path[sizeof(name.sun_path) - 1] = '\0';
    if ((fdCurrSock = socket (PF_UNIX, SOCK_STREAM, 0)) == -1)
        return -1;
    else if (bind (fdCurrSock, (struct sockaddr *) &name, 
            sizeof name.sun_family + 
            strlen (name.sun_path) + 1) == -1)
        return -1;
    else if (listen (fdCurrSock, 5) == -1)
        return -1;

    return fdCurrSock;
}

/**************************************
* handle an incoming socket message.
* constructs the new message object,
* adds it to the list of messages,
* and executes any commands.
*
* in: file descriptor of the socket
* out: number of bytes read. -1 if err
**************************************/
int process_msg (int p_fdIn)
{
    char msg[MSG_BUFF_SIZE];
    int intBytesRead, intNewPos;
    struct message msgNew; //message we're inserting
    struct tm* time_struct;
    long currtime;
    char * protocol_action; //pointer to protocol's action

    while( (intBytesRead=read(p_fdIn, msg, sizeof msg)) > 0)
        ;

    /************************************************
    * if not in originating desktop, add new message
    ************************************************/
    if((!MONITOR_DESKTOPS) || (get_desktop_num() != MSG_DESKTOP_NUM))
    {
        //set time of new msg
        currtime = time(0);
        time_struct = localtime(&currtime);
        if(time_struct)
            strftime (msgNew.time, MAX_TIME_LENGTH,
                TIME_FORMAT, time_struct);
        else
            strncat(msgNew.time, "00:00", 6);

        //set protocol of new message
        msgNew.protocol = match_protocol(global_config, msg);

        //add message
        intNewPos = addMessage(&msgNew);
        if(intNewPos < 0)
        {
            perror("Error adding message");
            return -1;
        }

        //fork the global command
        if((USE_GLOBAL_COMMAND) && (GLOBAL_COMMAND))
            execCommand(GLOBAL_COMMAND);

        //fork the protocol action
        protocol_action = PROTOCOL_ACTIONS[msgNew.protocol];
        if(protocol_action)
            execCommand(protocol_action);

        //tell main to update screen on next refresh
        global_screen_needs_updating = 1;
    }
    else
    {
        if(verbose) fprintf(stderr, "in originating desktop, ignoring message\n");
    }

    if(verbose) printAllMessages();
    return intBytesRead;
} //end process_msg

/******************************************
* adds a message to the list of messages.
* inserts message at element 0, and pushes
* the rest of the messages up one.
* also updates all blinking status.
*
* in: pointer to message data
* out: position of last message. -1 if err
******************************************/
int addMessage (struct message* p_msgNew)
{
    int intCurrPos;

    //if list isn't full, increment last message
    if(global_intLastMessage < MAX_MESSAGES-1)
        global_intLastMessage++;

    //shift all elts up one
    for(intCurrPos = global_intLastMessage; intCurrPos > 0; intCurrPos--)
    {
        global_messages[intCurrPos] = global_messages[intCurrPos-1];

        //shift blinking status as well, if no grouped icons
        if(BLINK_ENABLED && (!GROUPED_ICONS) && (intCurrPos < MESSAGES_PER_SCREEN))
        {
            if(verbose) fprintf(stderr, "global_blinks left[%i] = %i;\n", intCurrPos, global_blinks_left[intCurrPos-1]);
            global_blinks_left[intCurrPos] = global_blinks_left[intCurrPos-1];
        }
    }

    //initialize blink of new message
    if(BLINK_ENABLED)
    {
        if(GROUPED_ICONS)
        {
            if(verbose) fprintf(stderr, "global_blinks left[%i] = %i;\n", p_msgNew->protocol, NUMBER_OF_BLINKS*2);
            global_blinks_left[p_msgNew->protocol] = NUMBER_OF_BLINKS * 2;
        }
        else
        {
            if(verbose) fprintf(stderr, "global_blinks left[0] = %i;\n", NUMBER_OF_BLINKS*2);
            global_blinks_left[0] = NUMBER_OF_BLINKS * 2;
        }
    }

    //insert new message at elt 0
    global_messages[0].protocol = (*p_msgNew).protocol;
    strncpy(global_messages[0].time, (*p_msgNew).time, MAX_TIME_LENGTH);

    return global_intLastMessage;
}

/************************************************
* removes all messages of a particular protocol.
* calls removeMessage().
* 
* in: protocol number
************************************************/
void removeProtocol(int p_intProtocolNum)
{
    int intCurrPos;
    struct message *msgCurr;

    if(verbose) fprintf(stderr, "removing protocol #%i\n", p_intProtocolNum);
    for(intCurrPos = 0; intCurrPos <= global_intLastMessage; intCurrPos++)
    {
        msgCurr = &(global_messages[intCurrPos]);
        //remove message
        if(msgCurr->protocol == p_intProtocolNum)
        {
            if(removeMessage(intCurrPos) < -1)
            {
                fprintf(stderr, "Error: wmmsg can't remove"
                    " protocol #%i, located at position %i.\n",
                    p_intProtocolNum, intCurrPos);
                return;
            }

            intCurrPos--;
        }
    } //end for

    return;
}

/***********************************************
* removes a message from the list of messages.
* also updates all blinking status.
* note that it doesn't mess with display stuff.
*
* in: position of the message to delete
* out: 1 if successful, -1 if err
***********************************************/
int removeMessage (int p_intMessageNum)
{
    int intCurrPos;

    if(verbose) fprintf(stderr, "removing message %i\n", p_intMessageNum);
    //sanity check
    if((p_intMessageNum > global_intLastMessage) || (p_intMessageNum >= MAX_MESSAGES) || (p_intMessageNum < 0))
    {
        fprintf(stderr, "Error: wmmsg can't remove message"
            " #%i (last message located at: %i, max messages: %i).\n",
            p_intMessageNum, global_intLastMessage, MAX_MESSAGES);
        return -1;
    }

    //shift messages and set blink counts
    if(global_intLastMessage > 0)
    {
        //set blinking status if grouped icons
        if((BLINK_ENABLED) && (GROUPED_ICONS))
        {
            if(verbose) fprintf(stderr, "deleting: global_blinks_left[%i] = 0\n",
                global_messages[p_intMessageNum].protocol);
            if(global_blinks_left[global_messages[p_intMessageNum].protocol] > 0)
                global_blinks_left[global_messages[p_intMessageNum].protocol] = 0;
        }

        //shift down messages
        for(intCurrPos = p_intMessageNum; intCurrPos < global_intLastMessage; intCurrPos++)
        {
            global_messages[intCurrPos] = global_messages[intCurrPos+1];

            //if no grouped icons, shift blinks along with messages
            if((BLINK_ENABLED) && (!GROUPED_ICONS))
            {
                if(intCurrPos < MESSAGES_PER_SCREEN-1)
                    global_blinks_left[intCurrPos] = global_blinks_left[intCurrPos+1];
                else if(intCurrPos == MESSAGES_PER_SCREEN-1)
                    global_blinks_left[intCurrPos] = 0;
            }
        }
    }

    global_intLastMessage--;

    if(verbose) printAllMessages();
    return 1;
}

/********************************************
* removes all messages from list of messages
********************************************/
void removeAllMessages ()
{
    int i;

    if(BLINK_ENABLED)
    {
        if(GROUPED_ICONS)
        {
            for(i = 0; i < PROTOCOL_COUNT; i++)
                global_blinks_left[i] = 0;
        }
        else
        {
            for(i = 0; i < MESSAGES_PER_SCREEN; i++)
                global_blinks_left[i] = 0;
        }
    }

    global_intLastMessage = -1;
    if(verbose) printAllMessages();
    return;
}

/*****************************************
* prints messages to stdout for debugging
*****************************************/
void printAllMessages ()
{
    int intMsg;
    static char unknown[] = "???";
    char *strProtocol;

    printf("\n");
    if(global_intLastMessage == -1)
    {
        fprintf(stderr, " [no messages]\n");
    }
    else if(global_intLastMessage < -1)
    {
        fprintf(stderr, " [ERROR! no messages, but global_intLastMessage = %i]\n",
            global_intLastMessage);
    }
    else
    {
        for(intMsg=0; intMsg <= global_intLastMessage; intMsg++)
        {
            if(global_messages[intMsg].protocol == 0)
                strProtocol = unknown;
            else
                strProtocol = PROTOCOL_NAMES[global_messages[intMsg].protocol];

            if(GROUPED_ICONS)
            {
                if(BLINK_ENABLED)
                {
                    fprintf(stderr, " [%i] : %s  %s  (blinks: %i)\n",
                        intMsg, strProtocol, global_messages[intMsg].time,
                        global_blinks_left[global_messages[intMsg].protocol]);
                }
                else
                {
                    fprintf(stderr, " [%i] : %s  %s\n",
                        intMsg, strProtocol, global_messages[intMsg].time);
                }
            }
            else
            {
                if(BLINK_ENABLED)
                {
                    fprintf(stderr, " [%i] : %s  %s  (blinks: %i)\n",
                        intMsg, strProtocol, global_messages[intMsg].time,
                        (intMsg<MESSAGES_PER_SCREEN)? global_blinks_left[intMsg] : 0);
                }
                else
                {
                    fprintf(stderr, " [%i] : %s  %s\n",
                        intMsg, strProtocol, global_messages[intMsg].time);
                }
            }
        }
    }
}

/********************************************
* unlinks the socket and exits.
* only called on interrupt (ctrl-c, etc.)
* event. reads global_strSocketName variable
*
* in: signal number. not used
********************************************/
void clean_up (int p_intSignal)
{
    if(verbose) fprintf(stderr, "unlinking socket \"%s\"...", global_strSocketName);
    unlink(global_strSocketName);
    if(verbose) fprintf(stderr, " done.\n");
    _exit(0);
}


/**************************************************
* copied from wmpop3 for printing out a string.
* note: modified to only print a time (ie, no
* alphas).
*
* in: string to print, and destination coordinates
**************************************************/
void BlitString(char *p_strMsg, int p_intX, int p_intY)
{
    int   i;
    char  c;
    int   intX;

    intX = p_intX;
    for (i=0; p_strMsg[i]; i++)
    {
        //c = toupper(p_strMsg[i]);
        c = p_strMsg[i];
        if (c >= 'A' && c <= 'Z')
        {   /* its a letter */
            /* not dealing with chars now */
            /*
            c -= 'A';
            copyXPMArea(c * 6, 74, 6, 8, intX, p_intY);
            intX += 6;
            */
        }
        else if( c == ' ' ){
            c = 14;
            copyXPMArea(73, BLIT_XPM_Y, 6, 8, intX, p_intY);
            intX += 6;
        }
        else if( c == '/' ){
            c = 14;
            copyXPMArea(68, BLIT_XPM_Y, 6, 8, intX, p_intY);
            intX += 6;
        }
        else
        {   /* its a number or symbol */
            c -= '0';
            copyXPMArea(c * 6, BLIT_XPM_Y, 6, 8, intX, p_intY);
            intX += 6;
        }
    }
}

/**************************************
* gets the desktop currently in focus
*
* out: desktop number >= 0. if desktop
* isn't available, returns -1.
**************************************/
int get_desktop_num ()
{
    static Atom type_ret;
    static int format_ret;
    static unsigned long items_ret;
    static unsigned long after_ret;
    static unsigned char *prop_data = 0;

    (void) XGetWindowProperty (display, Root, global_win_workspace,
        0, 0x7fffffff, False, XA_CARDINAL, &type_ret,
        &format_ret, &items_ret, &after_ret, &prop_data);

    if(prop_data)
        return (int)(*prop_data);
    else
        return -1;
}

/*****************************************
* returns name of socket to use. located
* in tmp directory, and based on username
* (e.g., "/tmp/wmmsg_mike")
*
* out: name of socket, as char pointer
*****************************************/
char * getSocketName()
{
    struct passwd *pw;
    char * returner;

    returner = malloc(SOCKET_NAME_SIZE * sizeof(char));
    if(!returner) outOfMemory();

    pw = getpwuid(getuid());
    if(pw)
        snprintf(returner, SOCKET_NAME_SIZE, "%s/wmmsg_%s",
            P_tmpdir, pw->pw_name);
    else
        snprintf(returner, SOCKET_NAME_SIZE, "%s/wmmsg", P_tmpdir);

    return returner;
}

//prints out of memory error and exits
void outOfMemory()
{
    fprintf(stderr, "Error: wmmsg out of memory!\n");
    _exit(1);
}

