/*************************************************
* wmmsg_notify.c
*
* Sends a message over a UNIX socket to the
* listening wmmsg program. Takes arguments from
* the command line and sends them to the program.
*************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <sys/un.h>
#include <string.h>
#include <pwd.h>

#include "wmmsg.h"

/* constants */
#define MSG_BUFF_SIZE 1024

/* fun headers */
static int get_client_socket (char*);
static int send_msg (int, char*);
static char * getSocketName ();
static void displayUsage (void);

/***************************
* long command line options
***************************/
#ifdef HAVE_GETOPT_LONG
static struct option long_options[] =
{
    {   "help",       no_argument, NULL, 'h'},
    {"verbose",       no_argument, NULL, 'v'},
    {0, 0, 0, 0}
};
#else
void * long_options = NULL;
#endif

/******
* main
******/
int main (int argc, char *argv[])
{
    //name of socket (/tmp/...)
    char * strSockName;
    //file descriptor of socket
    int fdSocket;
    //msg to send
    char strMsg[MSG_BUFF_SIZE];
    //verbose mode (default off)
    char verbose = 0;
    //temps
    int intCurrOpt;
    int buffer_room_left = MSG_BUFF_SIZE - 1;

    //strMsg[MSG_BUFF_SIZE - 1] = '\0';

    /********************
    * parse command line
    ********************/
	while ((intCurrOpt = wmmsg_getopt(argc, argv, "hv", (void *)long_options, NULL)) != -1)
    {
        switch(intCurrOpt)
        {
            //-v: verbose mode
            case 'v':
                verbose = 1;
                break;
            //-h: output help
            case 'h':
                displayUsage();
                _exit(0);
            //other options: error
                case '?':
                //getopt_long already printed an error message
                fprintf(stderr, "Try `wmmsg_notify --help' for more information.\n");
                break;
        }
    }

    //must have at least 1 arg
    if(optind >= argc)
    {
        //fprintf(stderr, "Usage: %s <protocol>\n", argv[0]);
        fprintf(stderr, "Error: missing protocol\n");
        displayUsage();
        return -1;
    }

    //determine socket name
    if(verbose) fprintf(stderr, "getting socket name...");
    strSockName = getSocketName();
    if(verbose) fprintf(stderr, " \"%s\"\n", strSockName);

    //open socket
    if(verbose) fprintf(stderr, "opening socket...");
    fdSocket = get_client_socket(strSockName);
    free(strSockName);
    if(verbose) fprintf(stderr, " done\n");

    //if wmmsg isn't running, then just exit gracefully
    if(fdSocket == -1)
    {
        if (verbose) fprintf(stderr, "socket failed to open. I'm assuming that wmmsg isn't running, so I'm just exiting.\n");
        return -1;
    }

    //generate outgoing message
    if (verbose) fprintf(stderr, "generating outgoing message...");
    strncpy(strMsg, argv[optind], buffer_room_left);
    buffer_room_left -= strlen(argv[optind]);
    /*
    removed: we now only look at the first arg, because of gaim limitations
    for(optind++; optind<argc; optind++)
    {
        strncat(strMsg, " ", buffer_room_left--);
        strncat(strMsg, argv[optind], buffer_room_left);
        buffer_room_left -= strlen(argv[optind]);
    }
    */
    if (verbose) fprintf(stderr,
        " done.\nmessage: \"%s\"\nsending message over socket...", strMsg);

    //send message
    if (send_msg (fdSocket, strMsg) == -1)
    {
        if (verbose) fprintf(stderr, " write() returned -1!\n");
        perror ("Error sending message");
        return -1;
    }
    if (verbose) fprintf(stderr, " done.\nclosing socket...");

    //close socket
    if (close (fdSocket) == -1)
    {
        if (verbose) fprintf(stderr, " close() returned -1!\n");
        perror ("Error closing socket");
        return -1;
    }
    if (verbose) fprintf(stderr, " done. Exiting sucessfully.\n");
    return 1;
}

/**************************************************
* given the name of a socket, opens the socket for
* writing
* 
* in: name of socket
* out: file descriptor of opened socket. -1 if err
**************************************************/
int get_client_socket (char p_strSockName[])
{
    struct sockaddr_un name;
    int fdCurrSock;
    name.sun_family = AF_UNIX;
    strncpy (name.sun_path, p_strSockName, sizeof(name.sun_path)-1);
    name.sun_path[sizeof(name.sun_path)-1] = '\0';
    if ((fdCurrSock = 
        socket (PF_UNIX, SOCK_STREAM, 0)) == -1)
        return -1;
    else if (connect (fdCurrSock, (struct sockaddr *)&name, 
            sizeof name.sun_family 
            + strlen (name.sun_path) + 1) == -1)
        return -1;
    return fdCurrSock;
}

/********************************************
* writes a message to an open socket
*
* in: p_fdIn: file descriptor of open socket
*     p_strMsg: message to send
* out: number of bytes written, or -1 if err
********************************************/
int send_msg (int p_fdIn, char* p_strMsg)
{
    return (write(p_fdIn, p_strMsg, MSG_BUFF_SIZE));
}

/*****************************************
* returns name of socket to use. located
* in tmp directory, and based on username
* (e.g., "/tmp/wmmsg_mike")
*
* out: name of socket, as char pointer
*****************************************/
char * getSocketName()
{
    struct passwd *pw;
    char * returner;

    returner = malloc(SOCKET_NAME_SIZE * sizeof(char));
    if(!returner)
    {
        fprintf(stderr, "Error: wmmsg_notify out of memory!\n");
        _exit(1);
    }

    pw = getpwuid(getuid());
    if(pw)
        snprintf(returner, SOCKET_NAME_SIZE, "%s/wmmsg_%s",
            P_tmpdir, pw->pw_name);
    else
        snprintf(returner, SOCKET_NAME_SIZE, "%s/wmmsg", P_tmpdir);

    return returner;
}

/*************************
* gives command line help
*************************/
void displayUsage (void)
{
#ifdef HAVE_GETOPT_LONG
  fprintf (stderr, "Usage: wmmsg_notify [options] protocol\n\n");
  fprintf (stderr, "Options:\n");
  fprintf (stderr, "    -h, --help          display this text and exit\n");
  fprintf (stderr, "    -v, --verbose       verbose output\n");
  fprintf (stderr, "\n");
#else
  fprintf (stderr, "Usage: wmmsg_notify [options] protocol\n\n");
  fprintf (stderr, "Options:\n");
  fprintf (stderr, "    -h       display this text and exit\n");
  fprintf (stderr, "    -v       verbose output\n");
  fprintf (stderr, "\n");
#endif
}
